﻿include("functions_AD");
include("functions_target");
include("functions_basic");

/*********************************************************************************
**********************************************************************************
                                 FONCTIONS DE CALCUL 
**********************************************************************************
*********************************************************************************/

//limite l'apparition dans les logs des calculs de puissance en mode debug
function sayPower(cw, withEquip, target) {
	if (!isInTab(cw, functions_said_powers)) {
		if (isChip(cw))
			debug("puissance de " + getChipName(cw) + "(" + cw + "): " + getPower(cw, target)/getChipCost(cw));
		else {
			var costPT = getWeaponCost(cw);
			if (withEquip)
				costPT++;
			debug("puissance de " + getWeaponName(cw) + "(" + cw + "): " + getPower(cw, target)/costPT);
		}
		functions_said_powers = arrayAppend(functions_said_powers, cw);
	}
}

//retourne la puissance d'une puce ou d'une arme utilisée sur une cible
// les puces d'amélioration retournent actuellement le bonus qui sera appliqué à
// la meilleure arme/puce qui correspond, multiplié par le nombre de tour de l'effet.
//Si vous comptez tirer plusieurs fois, multipliez donc ce résultat!
//ne prend pas en compte le taux d'échec
//ne prend pas en compte les dégats/soins de zone, juste la puissance individuelle
//pour les armures, la plage de dégats stoppés est basée sur l'analyse du danger
//refresh_functions() est donc nécessaire au bon fonctionnement de cette évaluation
function getPower(cw, target)
{
	var relShield;
	var absShield;
	if (functions_simulation_isActive) {
		relShield = TA_GetRelativeShield(target);
		absShield = TA_GetAbsoluteShield(target);
	} else {
		relShield = getRelativeShield(target);
		absShield = getAbsoluteShield(target);
	}
	var effects;
	if (isChip(cw))	
		effects = getChipEffects(cw);
	else
		effects = getWeaponEffects(cw);
	var resultat = 0;
	for (var i in effects) {
		if (i[0] == EFFECT_DAMAGE || i[0] == EFFECT_POISON) { //l'effet de poison n'est pas réduit par l'armure!!
			var temp = mean(i[1],i[2]) * (1 + getForce() / 100) *
			(1 - relShield / 100) - absShield;
			temp = temp * (i[3]==0?1:i[3]); // fois le nombre de tours
			resultat = resultat + (temp>=0 ? temp: 0);
		} else if (i[0] == EFFECT_HEAL) {
			resultat = resultat + (1+getAgility()/100) * mean(i[1],i[2]) * i[3];
		} else if (i[0] == EFFECT_ABSOLUTE_SHIELD) {
			var result_shield = i[1] * (1+getAgility()/100);
			for (var j in functions_futurs_damage)
			{
				resultat = resultat + ((j-result_shield)>=0
							? result_shield:j);
			}
		} else if (i[0] == EFFECT_RELATIVE_SHIELD) {
			var result_shield = i[1] + getAgility()/50;
			for (var j in functions_futurs_damage)
			{
				resultat = resultat + j - 
					(j*result_shield/100);
			}
		} else if (i[0] == EFFECT_BUFF_AGILITY) { //avec CHIP_CURE codé en dur comme référence
			resultat = resultat + i[3] * (mean(35,43) * (1+(getAgility()+i[1])/100)-
				mean(35,43) * (1+(getAgility())/100));
		} else if (i[0] == EFFECT_BUFF_FORCE) {
			resultat = resultat + i[3] * (getMeanDamage(getWeapon()) * (1+(getForce()+i[1])/100)-
				getMeanDamage(getWeapon()) * (1+(getForce())/100));
		}
	}
	return resultat;
}

//retourne la puissance de l'arme/puce en fonction de son cout (basé sur getPower)
// le cout est augmenté de 1 si withEquip est à true (pour les armes)
function getPowerPerTP(cw, target, withEquip)
{
	if (isChip(cw)) {
		if (FUNCTION_PRINT_POWER)
			sayPower(cw, false, target);
		return getPower(cw, target)/getChipCost(cw);
	} else{
		if (FUNCTION_PRINT_POWER)
			sayPower(cw, withEquip, target);
		var costPT = getWeaponCost(cw);
		if (withEquip)
			costPT++;
		return getPower(cw, target)/costPT;
	}
}

//sans check de cible ni d'équipage, à n'utiliser que pour les puces non offensives
function getPowerPerTPF(cw) {
	if (!functions_simulation_isActive)
		return getPowerPerTP(cw, getLeek(), false);
	else
		return getPowerPerTP(cw, TA_ennemy[0], false);
}

//utilise tous vos points de tour pour utiliser la puce/arme sur la cible
function spendTPWith(weap_chip, target)
{
	if (weap_chip == null)
		return;
	if (isChip(weap_chip)) {
		if (!canUseChip(weap_chip, target))
			return;
		myUseChip(weap_chip, target);
		while(getTP() >= getChipCost(weap_chip) && canUseChip(weap_chip, target) && getChipCooldown(weap_chip) == 0) 
			myUseChip(weap_chip, target);
	} else {
		if (getWeapon() != weap_chip)
			setWeapon(weap_chip);
		if (!canUseWeapon(weap_chip, target))
			return;
		while(getTP() >= getWeaponCost(weap_chip) && canUseWeapon(weap_chip, target))
			useWeapon(target);
	}
	if (getTP() >= getChipCost(CHIP_SPARK))
		spendTPWith(CHIP_SPARK, target);
}

//évalue le cout en tp d'une suite de puces et d'armes, a utiliser dans cet ordre.
//on considère que l'arme actuellement équipée est "equiped"
function evalueCost(cw, equiped) {
	var lastWeap = equiped;
	var cost = 0;
	for (var i in cw) {
		if (isWeapon(i) && lastWeap != i) {
			lastWeap = i;
			cost++;
		}
		if (isWeapon(i))
			cost += getWeaponCost(i);
		else
			cost+= getChipCost(i);
	}
	return cost;
}

//retourne la force de frappe (effective) sur la cible avec la liste d'armes donnée
function evaluePower(cw, target) {
	var power = 0;
	for (var i in cw) { //on se fiche de savoir si équipé ou non, on veut juste la force brute
		power += getPowerPerTP(i, target, true) * getCWCost(i); // le prix par TP * TP
	}
	return power;
}

function getCostEquip(cws, cw, equiped) {
	if (isChip(cw))
		return 0;
	for (var i in cws) {
		if (isWeapon(i))
			equiped = i;
	}
	if (cw != equiped)
		return 1;
	return 0;
}

//construit la meilleure liste d'arme possible pour frapper votre cible.
//cw: liste de vos armes/puces
//tp: nombre de points de TP autorisés
//target: votre cible
//checkTerrain: si true, on ne prend que les armes qui seraient utilisables sur la cible
//equiped: votre arme actuellement équipée
//cout en opérations: bcp! a utiliser de manière tres limitée (genre juste 900k !)
function buildBetterWeapons_v2(cw, tp, target, checkTerrain, equiped) {
	var res = [[]];
	var temp = [];
	var added = true;
	var addedWeap;
	//1: on génère toutes les combinaisons d'armes possible
	while (added) {
		added = false;
		temp = [];
		for (var i in res) {
			addedWeap = false;
			for (var j in cw) {
				if (i[count(i)-1] >= j || count(i) == 0) {// on force le sens unique en comparant les valeurs d'armes (on évite des doublons miroirs)
					var costj = getCWCost(j);
					if (isWeapon(j) || (!isInRecovery(j) && !isInTab(j,i)) || getChipCooldown(j)==0) { //on check la récupération ET la multiplicité d'utilisation par tou
						var costEquip = getCostEquip(i, j, equiped);
						if (evalueCost(i, equiped) + costj + costEquip <= tp) {//si rajouter j reste en dessous des tp
							var copieI = i;
							push(copieI, j);
							temp[count(temp)] = copieI;//on le rajoute
							addedWeap = true;
							added = true;
						}
					}
				}
			}
			if (!addedWeap) // si on n'a rien pu rajouter, on garde la ligne
				temp[count(temp)] = i;
		}
		res = temp; // on garde les nouvelles valeurs.
	}
;	//2: on évalue et on garde la meilleur combinaison
	var damage = 0;
	var betterCW = null;
	for (var i in res) {
		if (!checkTerrain || getCellToUseWeapons(i, target) != null){ //checkTerrain 
			var power = evaluePower(i, target);
			if (power >= damage) {//si la combinaison est meilleure, on la garde
					betterCW = i;
					damage = power; //lourd en taches ca
			}
		}
	}
	if (FUNCTION_DEBUG_BUILDER_WEAPON)
		debug("la meilleure combinaison d'arme pour "+ tp + " TP est " + betterCW);
	return betterCW;
}

//DEPRECATED: utiliser buildBetterWeapons_v2!
//construit la meilleure combinaison d'arme possible pour toucher la cible en un nombre
//de points de tour donné.
//si checkTerrain vaut true, alors on vérifie que qu'on pourra atteindre une 
//case unique qui permet de les utiliser toutes, sinon non
function buildBetterWeapons(cw, tp, target, checkTerrain)
{
	if (cw == null || count(cw) ==0)
		return [];
	var res = []; var index = 0;
	var equiped_w;
	var tempTP = tp;
	var weap_chips = cw;
	var bestW = cw[0];
	var lastWeap = getWeapon();
	while (tempTP >= getCWCost(getCheapestCW(weap_chips)))
	{
		for (var i in weap_chips)
		{
			//on vérifie si
			//on a trouvé une cw plus puissante
			if (getPowerPerTP(bestW, target, (lastWeap == i)) < 
				getPowerPerTP(i, target, (lastWeap == i)))
				//Si le cout convient
				if ((getCWCost(i)+((lastWeap==i || isChip(i))?0:1))<=tempTP)
					//si la puce n'est pas en cours de recharge
					if (isWeapon(i) || !isInRecovery(i))
						//si demandé, que le terrain est valable
						if (!checkTerrain || getCellToUseWeapons(arrayAppend(res, i), target) != null)
							bestW = i;
		}
		if ((getCWCost(bestW)+((lastWeap==bestW || isChip(bestW))?0:1))<=tempTP) {
			res[index] = bestW;
			tempTP = tempTP - getCWCost(bestW) ;
			if (lastWeap != bestW && isWeapon(bestW))
				tempTP--;
			index++;
		} else
			tempTP = -1;
		if (isWeapon(bestW))
			lastWeap = bestW;
	}
	if (FUNCTION_DEBUG_BUILDER_WEAPON)
		debug("la meilleure combinaison d'arme pour "+ tp + " TP est " + res);
	return res;
}

//retourne toutes les cases ou on peut utiliser toutes les armes (puce) envoyées en 
//paramètre et les marque en noir sur le terrain. Si aucune, renvoie null
function getCellToUseWeapons(cw, target)
{
	if (functions_simulation_isActive) {
		debugE("Vous avez demandé une case commune à des armes alors que la simulation est active");
		return null;
	}
	var cells= [];
	var cellsres = [];
	if (count(cw) == 0)
		return getCell();
	for (var i in cw) {
		if (isChip(i))
			cells[count(cells)] = getCellsToUseChip(i, target,getLeek());
		else
			cells[count(cells)] = getCellsToUseWeapon(i, target,getLeek());
	}
	for (var i in cells[0]) {
		for (var j in cells) {
			var isIn = true;
			if (!isInTab(i, j))
				isIn = false;
			if (isIn)
				cellsres[count(cellsres)] = i;
		}
	}
	if (FUNCTION_MARK_ALLOWED_CELLS) {
		mark(cellsres, FUNCTION_MARK_ALLOWED_CELLS_COLOR);
	}
	if (FUNCTION_MARK_TARGET_CELL)
		mark(getNearlyCell(cellsres), FUNCTION_MARK_TARGET_CELL_COLOR);
	return getNearlyCell(cellsres);
}

//utilise toutes les armes envoyées en paramètre sur la cible (doublons possibles)
function useWeaponsOnTarget(weapons, target)
{
	for (var i in weapons) {
		if (isWeapon(i)) {
			if (getWeapon() != i)	
				setWeapon(i);
			useWeapon(target);
		} else
			myUseChip(i, target);
	}
}

//peut on utiliser l'arme cw depuis cell sur la cellule ct?
function canUseCWFrom(cw, cell, ct)
{
	if (isChip(cw)) {
		if (isInlineChip(cw)) {
			if (getCellX(cell) == getCellX(ct) && getCellY(cell) == getCellY(ct))
				if (getChipMinScope(cw) <= getDistance(cell, ct) && getDistance(cell, ct) <= getChipMaxScope(cw))
					return true;
		} else if (cw != CHIP_SPARK) {
			if (getPathLength(cell, ct) <= getChipMaxScope(cw) && getPathLength(cell, ct) >= getChipMinScope(cw)) {
				//vérifier si il n'y a pas d'obstacle au moyen de la longueur du chemin:
				var dist = abs(getCellX(cell) - getCellX(ct)) + abs(getCellY(cell) - getCellY(ct)) + 1;
				if (getPathLength(cell, ct) == dist)
					return true;
			}
		} else {//dans le cas de chip_spark
			var dist = abs(getCellX(cell) - getCellX(ct)) + abs(getCellY(cell) - getCellY(ct)) + 1;
			if (dist <= 10)
				return true;
		}
	} else if (isWeapon(cw)) {
		if (isInlineWeapon(cw)) {
			if (getCellX(cell) == getCellX(ct) && getCellY(cell) == getCellY(ct))
				if (getWeaponMinScope(cw) <= getDistance(cell, ct) && getDistance(cell, ct) <= getWeaponMaxScope(cw))
					return true;
		} else {
			if (getPathLength(cell, ct) <= getWeaponMaxScope(cw) && getPathLength(cell, ct) >= getWeaponMinScope(cw)) {
				//vérifier si il n'y a pas d'obstacle au moyen de la longueur du chemin:
				var dist = abs(getCellX(cell) - getCellX(ct)) + abs(getCellY(cell) - getCellY(ct)) + 1;
				if (getPathLength(cell, ct) == dist)
					return true;
			}
		}
	}
	return false;
}

//enlève des armes envoyées en paramètre toutes les armes qu'il sera impossible d'utiliser
//, c'est à dire toutes les armes qui ne pourraient toucher personne, même en se déplacant
function removeUselessCW(cw, target) {
	var res = [];
	for (var i in cw) {
		var cells;
		if (isWeapon(i))
			cells = getCellsToUseWeapon(i, target);
		else 
			cells = getCellsToUseChip(i, target);
		if (getMP() >= getPathLength(getCell(),getNearlyCell(cells)))
			res = arrayAppend(res, i);
	}
	return res;
}

//renvoie la meilleurs puces de toutes celles qui ont au moins un des effets désiré 
//(les puces sont testées sur soi-meme)
function bestAChipWithEffect(effects) {
	var chips = [];
	for (var i in effects)
		chips = arrayConcat(chips, getChipsWithEffect(i));
	if (count(chips) == 0)
		return null;
	var best = chips[0];
	for (var i in chips) {
		if (getPowerPerTPF(i) >= getPowerPerTPF(best) || !myCanUseChip(best, getLeek()))
			if (myCanUseChip(i, getLeek()))
				best = i;
	}
	if (myCanUseChip(best, getLeek()))
		return best;
	else
		return null;
}

//renvoie le nombre de points de tours pour que ceux-ci aient le maximum d'effets!
//il vaut parfois mieux tirer une fois avec le laser et dépenser le reste en autre chose
//que de tirer avec le pistolet ou spark par apres...
//demande la liste des armes, une cible et un intervalle de calcul
//retourne un tableau avec comme première case le cout optimal en TP
// et comme deuxième case un tableau avec la liste des armes déjà calculée :p
//(calcul avec buildBetterWeapon_v2, checkTerrain, et arme équipée).
function getBetterTPUsage(cw, target, min, max) {
	var tab = [];
	for (var i=min; i<=max; i++) {
		tab[i-min] = buildBetterWeapons_v2(cw, i, target, true, getWeapon());
	}
	var meanDamage = 0;
	var indexFound = -1;
	for (var i=0; i<count(tab); i++) {
		if (evaluePower(tab[i][1], target) / tab[i][0] > meanDamage) {
			indexFound = i;
			meanDamage = evaluePower(tab[i][1], target) / tab[i][0];
		}
	}	
	return [indexFound + min, tab[indexFound][1]]